//---------------------------------------------------------------------
//		̑̊֐
//---------------------------------------------------------------------
#include <windows.h>
#include <mmsystem.h>
#include <emmintrin.h>

#include "yv12cnv.h"
#include "mms_util.h"
#include "mms.h"

//---------------------------------------------------------------------
//	Rs[p֐isô݃oCgPʁAȊO̓sNZPʁj
//---------------------------------------------------------------------
static void copy_line(BYTE *pOut, const BYTE *pIn,
					  UINT bpp, UINT h, UINT w, UINT si, UINT so);
static void copy_line_r(BYTE *pOut, const BYTE *pIn,
						UINT bpp, UINT h, UINT w, UINT si, UINT so);
static void copy_line_sse2(BYTE *pOut, const BYTE *pIn,
						   UINT bpp, UINT h, UINT w, UINT si, UINT so);
static void copy_line_r_sse2(BYTE *pOut, const BYTE *pIn,
							 UINT bpp, UINT h, UINT w, UINT si, UINT so);
static void copy_null(BYTE *pOut, const BYTE *pIn,
					  UINT bpp, UINT h, UINT w, UINT si, UINT so);
static bool isSSE2available();
inline void CopyMemorySSE2(__m128i *dst, const __m128i *src, size_t len);

//---------------------------------------------------------------------
// SSE2ɑΉĂ邩`FbN
//---------------------------------------------------------------------
bool isSSE2available()
{
	unsigned long r_ecx,r_edx;

	// CPUIDɂ`FbN
	__asm {
		mov eax, 1
		cpuid
		mov r_edx, edx
		mov r_ecx, ecx
	}

	// MMX  0x0800000 edx
	// SSE  0x2000000 edx
	// SSE2 0x4000000 edx
	// SSE3 0x0000001 ecx

	if(!(r_edx & 0x4000000U))
		return false;

	// OST|[gĂ邩
	__try {
		__asm xorps xmm0, xmm0
	}

	__except(EXCEPTION_EXECUTE_HANDLER){
	if (_exception_code()==STATUS_ILLEGAL_INSTRUCTION)
		/* SSE not supported */
		return false;
	}
	/* SSE are supported by OS */
	return true;
}

//---------------------------------------------------------------------
// SSE2pŃRs[֐i16byte align16̔{oCgOj
//---------------------------------------------------------------------
void CopyMemorySSE2(__m128i *dst, const __m128i *src, size_t len)
{
	__m128i x0,x1,x2,x3;
#if 0
	for( ; len >= 64U; len -= 64U){
		x0 = src[0];
		x1 = src[1];
		x2 = src[2];
		x3 = src[3];
		src += 4;
		_mm_stream_si128(dst    , x0);
		_mm_stream_si128(dst + 1, x1);
		_mm_stream_si128(dst + 2, x2);
		_mm_stream_si128(dst + 3, x3);
		dst += 4;
	}
	for( ; len >= 16U; len -= 16U){
		_mm_stream_si128(dst++, *src++);
	}
#else
	for( ; len >= 64U; len -= 64U){
		x0 = src[0];
		x1 = src[1];
		x2 = src[2];
		x3 = src[3];
		src += 4;
		dst[0] = x0;
		dst[1] = x1;
		dst[2] = x2;
		dst[3] = x3;
		dst += 4;
	}
	for( ; len >= 16U; len -= 16U){
		*dst++ = *src++;
	}
#endif
}


//---------------------------------------------------------------------
//		Rs[pNX
//---------------------------------------------------------------------
ImgCopy::ImgCopy(DWORD fmt, bool mmx, bool sse2,
				 UINT swIn, UINT Height, UINT swOut, UINT Width)
 : _fmt(fmt), _swIn(swIn), _Height(Height), _swOut(swOut), _Width(Width)
{
	SetCopyFunc(mmx, sse2);
}

void ImgCopy::Copy(BYTE *dst, const BYTE *src)
{
	//dst 16byte align O
	if(0 == (reinterpret_cast<ULONG_PTR>(src) & 0xfU)){
		_cpfunc(dst, src, _bpp, _Height, _Width, _swIn, _swOut);
	}else{
		_cpfunc_na(dst, src, _bpp, _Height, _Width, _swIn, _swOut);
	}
}

void ImgCopy::SetCopyFunc(bool mmx, bool sse2)
{
	sse2 = isSSE2available() && sse2;

	switch (_fmt) {
		case MM_FMT_RGB24:
			_fcc = 0;
			_bpp = 3;
			_buf_cnt_mul = 2;
			SetLineCopyFunc(false, sse2);
			break;
		case MM_FMT_RGB24R:
			_fcc = 0;
			_bpp = 3;
			_buf_cnt_mul = 2;
			SetLineCopyFunc(true, sse2);
			break;
		case MM_FMT_YUY2:
			_fcc = MAKEFOURCC('Y','U','Y','2');
			_bpp = 2;
			_buf_cnt_mul = 3;
			SetLineCopyFunc(false, sse2);
			break;
		case MM_FMT_YV12:
			_fcc = MAKEFOURCC('Y','U','Y','2');
			_bpp = 2;
			_buf_cnt_mul = 3;
			SetYV12CnvFunc(false, mmx, sse2);
			break;
		case MM_FMT_YV12I:
			_fcc = MAKEFOURCC('Y','U','Y','2');
			_bpp = 2;
			_buf_cnt_mul = 3;
			SetYV12CnvFunc(true, mmx, sse2);
			break;
		case MM_FMT_YC48:
			_fcc = MAKEFOURCC('Y','C','4','8');
			_bpp = 6;
			_buf_cnt_mul = 1;
			SetLineCopyFunc(false, sse2);
			break;
		default:
			_fcc = 0;
			_bpp = 1;
			_buf_cnt_mul = 1;
			_cpfunc = _cpfunc_na = copy_null;
	}
}

void ImgCopy::SetLineCopyFunc(bool bu, bool sse2)
{
	UINT a = ((_bpp * _Width) | _swOut) & 15U;

	if(a == 0 && sse2){
		_cpfunc = bu ? copy_line_r_sse2 : copy_line_sse2;
	}else{
		_cpfunc = bu ? copy_line_r : copy_line;
	}
	_cpfunc_na = bu ? copy_line_r : copy_line;
}

void ImgCopy::SetYV12CnvFunc(bool ires, bool mmx, bool sse2)
{
	UINT a = (_Width | _swOut) & 15U;
	UINT b = _Width & 7U;

	if(0 == a && sse2){
		_cpfunc = ires ? sse2_iq_yuy2_2_yv12 : sse2_yuy2_2_yv12;
		_cpfunc_na = ires ? mmx_iq_yuy2_2_yv12 : mmx_yuy2_2_yv12;
	}else if(0 == b && mmx){
		_cpfunc = _cpfunc_na = ires ? mmx_iq_yuy2_2_yv12 : mmx_yuy2_2_yv12;
	}else{
		_cpfunc = _cpfunc_na = ires ? iq_yuy2_2_yv12 : yuy2_2_yv12;
	}
}

long ImgCopy::GetPitch(long align, long w, long bpp)
{
	if(align == 0) align = 1;
	
	long s = w * bpp;
	long r = s % align;

	s += r ? (align - r) : 0;

	/* sNZPʂŊ؂ȂƂ0Ԃ */
	return (s % bpp) ? 0 : s;
}

bool ImgCopy::CheckFormat(DWORD fmt, long h, long w, long &pitch)
{
	if(fmt < MM_FMT_RGB24 || fmt > MM_FMT_YC48){
		// ΉtH[}bg
		return false;
	}

 	/* XgCh̐ݒ */
	long s;

	switch (fmt) {
		case MM_FMT_RGB24:
		case MM_FMT_RGB24R:
			s = GetPitch(pitch, w, 3);
			break;
		case MM_FMT_YUY2:
			s = GetPitch(pitch, w, 2);
			/* 2̔{̂ */
			s = (w % 2) ? 0 : s;
			break;
		case MM_FMT_YV12:
		case MM_FMT_YV12I:
			s = GetPitch(pitch, w, 1);
			/* c2̔{̂ */
			s = (w % 2) ? 0 : s;
			s = (h % 2) ? 0 : s;
			break;
		case MM_FMT_YC48:
			s = GetPitch(pitch, w, 6);
			break;
		default:
			s = 0;
	}

	if(s == 0){
		return false;
	}

	/* XgChZbg */
	pitch = s;

	return true;
}

//---------------------------------------------------------------------
//		Rs[p֐
//---------------------------------------------------------------------
/* CPʂŃRs[ */
void copy_line(BYTE *pOut, const BYTE *pIn,
			   UINT bpp, UINT h, UINT w, UINT si, UINT so)
{
	si *= bpp;
	w  *= bpp;

	for(UINT i=0; i<h; ++i){
		CopyMemory(pOut, pIn, w);
		pOut += so;
		pIn  += si;
	}
}
					  
/* CPʂŃRs[/t */
void copy_line_r(BYTE *pOut, const BYTE *pIn,
				 UINT bpp, UINT h, UINT w, UINT si, UINT so)
{
	si *= bpp;
	w  *= bpp;

	pIn += si * (h - 1);

	for(UINT i=0; i<h; ++i){
		CopyMemory(pOut, pIn, w);
		pOut += so;
		pIn  -= si;
	}
}

/* CPʂŃRs[ SSE2 i16byte align16̔{oCgOj*/
void copy_line_sse2(BYTE *pOut, const BYTE *pIn,
					UINT bpp, UINT h, UINT w, UINT si, UINT so)
{
	si *= bpp;
	w  *= bpp;

	for(UINT i=0; i<h; ++i){
		CopyMemorySSE2((__m128i*)pOut, (const __m128i*)pIn, w);
		pOut += so;
		pIn  += si;
	}
}

/* CPʂŃRs[/t SSE2 i16byte align16̔{oCgOj*/
void copy_line_r_sse2(BYTE *pOut, const BYTE *pIn,
					  UINT bpp, UINT h, UINT w, UINT si, UINT so)
{
	si *= bpp;
	w  *= bpp;

	pIn += si * (h - 1);

	for(UINT i=0; i<h; ++i){
		CopyMemorySSE2((__m128i*)pOut, (const __m128i*)pIn, w);
		pOut += so;
		pIn  -= si;
	}
}

/* Ȃ */
void copy_null(BYTE*, const BYTE*, UINT, UINT, UINT, UINT, UINT)
{
}

