//---------------------------------------------------------------------
//		bZ[WV[o
//---------------------------------------------------------------------
#include <windows.h>
#include <process.h>

#include "mms.h"
#include "msg_wnd.h"

//MsgRecv
MsgRecv::MsgRecv(HINSTANCE dll, MmSrv *mms)
 : _dll(dll), _wnd(NULL), _mms(mms)
{
	WNDCLASSEX wc = { sizeof(WNDCLASSEX), };
	
	wc.lpfnWndProc   = msgProc;
	wc.hInstance     = _dll;
	wc.lpszClassName = TEXT("mm_srv_msg");

	_atm = RegisterClassEx(&wc);
}

MsgRecv::~MsgRecv()
{
	DestroyMsgWnd();

	if(_atm){
		UnregisterClass(MAKEINTATOM(_atm), _dll);
	}
}

bool MsgRecv::StartReceiving()
{
	if(!_wnd){
		return CreateMsgWnd();
	}

	return true;
}

bool MsgRecv::StopReceiving()
{
	return DestroyMsgWnd();
}

void MsgRecv::PeekMsg()
{
	MSG msg;

	while(PeekMessage(&msg, _wnd, 0, 0, PM_REMOVE)){
		DispatchMessage(&msg);
		DPRINT(TEXT("MsgRecv::PeekMsg(%lu)\n"), msg.message);
	}
}

// private

bool MsgRecv::CreateMsgWnd()
{
	if(_atm == 0 || _wnd){
		return false;
	}

	//crreate message-only window
	_wnd = CreateWindow(MAKEINTATOM(_atm),
		NULL, 0, 0, 0, 0, 0,
		HWND_MESSAGE, NULL, _dll, this);

	return _wnd ? true : false;
}

bool MsgRecv::DestroyMsgWnd()
{
	HWND tmp = static_cast<HWND>(InterlockedExchangePointer(&_wnd, NULL));

	if(tmp){
		DestroyWindow(tmp);
	}

	return true;
}

void MsgRecv::OnWindowCreate(HWND hwnd)
{
	SetWindowLongPtr(hwnd, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(this));
}

LRESULT CALLBACK MsgRecv::msgProc(HWND h, UINT u, WPARAM w, LPARAM l)
{
	MsgRecv *mr = reinterpret_cast<MsgRecv*>(GetWindowLongPtr(h, GWLP_USERDATA));

	switch (u) 
	{ 
	case MM_MSG_OPEN:
		if(mr == NULL) return 0;
		return mr->_mms->ReqOpen(static_cast<DWORD>(w), static_cast<DWORD>(l));
	case MM_MSG_GET:
		if(mr == NULL) return 0;
		return mr->_mms->ReqGetFrameOffs(static_cast<DWORD>(l));
	case MM_MSG_HIT:
		if(mr == NULL) return 0;
		return mr->_mms->ReqHitFrame(static_cast<DWORD>(l));
	case MM_MSG_CLOSE:
		if(mr == NULL) return 0;
		return mr->_mms->ReqClose();
	case WM_CREATE:
		reinterpret_cast<MsgRecv*>(
			reinterpret_cast<CREATESTRUCT*>(l)->lpCreateParams)->OnWindowCreate(h);
		return 0;
	} 

	return DefWindowProc(h, u, w, l);
}
